idelayoutstackheader {
  min-height: 26px;
}
idelayoutstackheader button {
  border: none;
  border-radius: 0;
  background: transparent;
  box-shadow: none;
  padding: 0;
  margin: 0;
}
idelayoutstackheader button:disabled {
  color: alpha(currentColor, 0.25);
}
idelayoutstackheader button:not(:disabled) image {
  opacity: 0.55;
}
idelayoutstackheader button:checked image,
idelayoutstackheader button:not(:disabled):hover image {
  opacity: 1;
}
idelayoutstackheader button:checked,
idelayoutstackheader button:hover {
  background: shade(@theme_bg_color, 0.9);
}
idelayoutstackheader button:active {
  background: shade(@theme_bg_color, 0.85);
}
idelayoutstackheader > button {
  padding-left: 12px;
  padding-right: 12px;
}
idelayoutstackheader * button:first-child > image,
idelayoutstackheader * button:last-child:dir(rtl) > image {
  padding-left: 12px;
  padding-right: 10px;
}
idelayoutstackheader * button:first-child:dir(rtl) > image,
idelayoutstackheader * button:last-child > image {
  padding-right: 12px;
  padding-left: 9px;
}
idelayoutgridcolumn.handle,
idelayoutgrid.handle {
  border: 1px solid @borders;
}
popover.title-popover list row {
  padding: 1px 0;
}
popover.title-popover scrolledwindow {
  min-width: 300px;
}
popover.title-popover list {
  background: transparent;
}
popover.title-popover list row box > image:first-child {
  opacity: 0.55;
}
popover.title-popover list row box > image:first-child {
  margin: 6px 8px;
  min-height: 16px;
  min-width: 16px;
}
popover.title-popover list row button:last-child {
  background: none;
  box-shadow: none;
  border: none;
  outline: none;
  padding: 0;
  margin: 0 6px;
  color: @theme_fg_color;
  opacity: 0.55;
}
popover.title-popover list row button:last-child:hover {
  opacity: 1;
}
popover.title-popover list row button image {
  min-height: 16px;
  min-width: 16px;
}
popover.title-popover list row button image {
  color: @theme_unfocused_fg_color;
}
